// SPDX-License-Identifier: BSD-3-Clause

#include "osmem.h"
#include <unistd.h>
#include <sys/mman.h>

#define MMAP_THRESHOLD   (128 * 1024)
#define page_size        4096
#define PROT_READ        0x1
#define PROT_WRITE       0x2
#define MAP_PRIVATE      0x02
#define MAP_ANONYMOUS    0x20
#define MAP_FAILED       ((void *) -1)

/* Structure to hold memory block metadata */
struct block_meta {
	size_t size;
	int status;
	struct block_meta *prev;
	struct block_meta *next;
};

/* Block metadata status values */
#define STATUS_FREE   0
#define STATUS_ALLOC  1
#define STATUS_MAPPED 2

struct block_meta *head;
void *heap;
int heap_initialized;

/* Allign the size to 8 bytes */
size_t allign_eight(size_t size)
{
	size_t rest = size % 8;

	if (rest == 0)
		return size;
	else
		return size + 8 - rest;
}

/* Initializing the heap */
void InitializeHeap(void)
{
	if (heap == NULL) {
		heap = sbrk(MMAP_THRESHOLD);
		if (heap == (void *)-1)
			return;

		/* Create a new block_meta structure at the start of the heap */
		struct block_meta *initial_block = (struct block_meta *)heap;

		initial_block->status = STATUS_FREE;
		initial_block->size = MMAP_THRESHOLD - sizeof(struct block_meta);
		initial_block->next = NULL;
		initial_block->prev = NULL;

		// Insert the initial block into the linked list
		if (head == NULL) {
			head = initial_block;
		} else {
			/* If head is not NULL, add it to the end of the list */

			struct block_meta *curr = head;

			while (curr->next != NULL)
				curr = curr->next;

			curr->next = initial_block;
			initial_block->prev = curr;
		}
	}
}

/* Finding the best block for the memory */
struct block_meta *find_best_size(size_t size)
{
	struct block_meta *block_memorie = head;
	struct block_meta *best_block = NULL;

	while (block_memorie) {
		if (block_memorie->status == STATUS_FREE && block_memorie->size >= size) {
			if (best_block == NULL)
				best_block = block_memorie;
			else if (block_memorie->size < best_block->size)
				best_block = block_memorie;
		}
		block_memorie = block_memorie->next;
	}

	return best_block;
}

struct block_meta *req_memorie(size_t size, size_t threshold)
{
	struct block_meta *block = NULL;

	size = allign_eight(size);

	if (size + sizeof(struct block_meta) < threshold) {
		/* Using the heap + sbrk if size < threshold bytes */

		/* Initializing the heap */
		if (!heap_initialized) {
			InitializeHeap();

			heap_initialized = 1;
		}

		/* Finding the best place in heap for the memory */
		block = find_best_size(size);
		if (block) {
			/* Checking with how much the block is bigger and splitting it,
			 * also ensuring the remaining size is enough for block_meta +1 and alligned
			 */

			if (block->size >= size + sizeof(struct block_meta) + 8) {
				/* Splitting the block */

				struct block_meta *new_block = (struct block_meta *)((char *)block + size + sizeof(struct block_meta));

				new_block->status = STATUS_FREE;
				new_block->size = block->size - size - sizeof(struct block_meta);
				new_block->next = block->next;
				new_block->prev = block;

				if (block->next != NULL)
					new_block->next->prev = new_block;

				block->size = size;
				block->next = new_block;
			}
			block->status = STATUS_ALLOC;

		} else {
			/* No good block, extending the heap */
			/* If the last block is free, expand it
			 * else, extend the heap
			 */

			struct block_meta *last_block = head;

			/* Ifthere is memory mapped it means we reached end of the heap */
			while (last_block->next && last_block->next->status != STATUS_MAPPED)
				last_block = last_block->next;

			/* Last block is free */
			if (last_block->status == STATUS_FREE) {
				size_t rest = size - last_block->size;

				/* Expanding the heap */
				sbrk(rest);
				last_block->size += rest;
				last_block->status = STATUS_ALLOC;
				block = last_block;

			} else {
				/* Extending the heap by a new block*/
				block = sbrk(size + sizeof(struct block_meta));
				if (block == (void *)-1)
					return NULL;

				block->status = STATUS_ALLOC;
				block->size = size;
				block->next = NULL;
				block->prev = last_block;
				last_block->next = block;
			}
		}

	} else {
		/* Using mmap if the size is > threshold kb */

		block = mmap(NULL, size + sizeof(struct block_meta), PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
		if (block == MAP_FAILED)
			return NULL;

		block->status = STATUS_MAPPED;
		block->size = size;
		block->next = NULL;
		block->prev = NULL;

		/* Add the mmaped block to the heap */
		if (head == NULL && heap_initialized == 0) {
			head = block;
		} else {
			struct block_meta *curr = head;

			while (curr->next)
				curr = curr->next;

			curr->next = block;
			block->prev = curr;
		}
	}

	return block;
}

void *os_malloc(size_t size)
{
	if (size == 0)
		return NULL;

	/* Alligin the size to 8 bytes */
	size = allign_eight(size);

	/* Request memory */
	struct block_meta *block_memorie = req_memorie(size, MMAP_THRESHOLD);

	if (block_memorie == NULL)
		return NULL;

	return (void *)((char *)block_memorie + sizeof(struct block_meta));
}

void os_free(void *ptr)
{
	if (ptr == NULL)
		return;

	/* Going to the meta-data of the block of memory */
	struct block_meta *mem = (struct block_meta *)((char *)ptr - sizeof(struct block_meta));

	if (mem->status == STATUS_FREE)
		return;

	if (mem->status == STATUS_ALLOC) {
		/* If it is < 128kb mark it as free */
		mem->status = STATUS_FREE;

		/* If the previous block is also free, unite them */
		while (mem->prev != NULL && mem->prev->status == STATUS_FREE) {
			mem->prev->size += mem->size + sizeof(struct block_meta);
			mem->prev->next = mem->next;
			if (mem->next != NULL)
				mem->next->prev = mem->prev;

			mem = mem->prev;
		}

		/* If the next block is also free, unite them */
		while (mem->next != NULL && mem->next->status == STATUS_FREE) {
			mem->size += mem->next->size + sizeof(struct block_meta);
			mem->next = mem->next->next;
			if (mem->next != NULL)
				mem->next->prev = mem;
		}

	} else if (mem->status == STATUS_MAPPED) {
		/* If it is > 128kb delete it from the list and munmap that shi */

		if (mem->prev != NULL) {
			mem->prev->next = mem->next;
		} else {
			if (mem == head) {
				head = mem->next;
				if (head != NULL)
					head->prev = NULL;
			}
		}

		if (mem->next != NULL)
			mem->next->prev = mem->prev;

		munmap(mem, mem->size + sizeof(struct block_meta));
	}
}

void *memset(void *source, int value, size_t num)
{
	char *byte = (char *)source;

	for (size_t i = 0; i < num; i++) {
		*byte = (char)value;
		byte++;
	}

	return source;
}

void *os_calloc(size_t nmemb, size_t size)
{
	size_t total_size = nmemb * size;

	if (total_size == 0)
		return NULL;

	/* Request memory */
	struct block_meta *block_memorie = req_memorie(total_size, page_size);

	if (block_memorie == NULL)
		return NULL;

	void *ptr = (void *)((char *)block_memorie + sizeof(struct block_meta));

	if (block_memorie->status != STATUS_MAPPED)
		memset(ptr, 0, total_size);

	return ptr;
}

void *memcpy(void *dest, const void *src, size_t n)
{
	char *d = (char *)dest;
	const char *s = (const char *)src;

	for (size_t i = 0; i < n; i++)
		d[i] = s[i];

	return dest;
}

void *os_realloc(void *ptr, size_t size)
{
	if (ptr == NULL)
		return os_malloc(size);

	if (size == 0) {
		os_free(ptr);
		return NULL;
	}

	/* Going to the meta-data of the block of memory */
	struct block_meta *mem = (struct block_meta *)((char *)ptr - sizeof(struct block_meta));

	if (mem->status == STATUS_FREE)
		/* Return NULL if the block is already free */
		return NULL;

	/* Allign the size to 8 bytes */
	size = allign_eight(size);

	if (mem->status == STATUS_ALLOC) {
		if (size <= mem->size) {
			/* Size is smaller, reducing from the block */

			size_t remaining_size = mem->size - size;

			if (remaining_size >= sizeof(struct block_meta) + 8) {
				/* Splitting the block */

				struct block_meta *new_block = (struct block_meta *)((char *)mem + sizeof(struct block_meta) + size);

				new_block->status = STATUS_FREE;
				new_block->size = remaining_size - sizeof(struct block_meta);
				new_block->next = mem->next;
				new_block->prev = mem;

				if (mem->next != NULL)
					mem->next->prev = new_block;

				mem->size = size;
				mem->next = new_block;
			}

			/* The original pointer, resized */
			return ptr;

		} else if (size > mem->size) {
			/* Attempting to expand the block in place */
			/* First, we search rigth to see if the dimmension is enough
			 * and the blocks remain coalesced
			 */

			if (size >= MMAP_THRESHOLD) {
				/* If the reallocation has a size bigger than the threshold, we mmap that shi */

				void *new_ptr = os_malloc(size);

				if (new_ptr == NULL)
					return NULL;

				size_t copy;

				if (mem->size < size)
					copy = mem->size;
				else
					copy = size;


				memcpy(new_ptr, ptr, copy);
				os_free(ptr);

				return new_ptr;
			}

			size_t total_size = mem->size;
			struct block_meta *next_block = mem->next;

			/* Adding the total free size + what we have now */
			while (next_block != NULL && next_block->status == STATUS_FREE) {
				total_size += next_block->size + sizeof(struct block_meta);
				next_block = next_block->next;
			}

			if (total_size >= size) {
				/* We have enough space, we will unite the blocks */

				/* Coalesce all the blocks in use */
				struct block_meta *current_block = mem;

				while (current_block->next && current_block->next->status == STATUS_FREE) {
					current_block->size += current_block->next->size + sizeof(struct block_meta);
					current_block->next = current_block->next->next;

					if (current_block->next != NULL)
						current_block->next->prev = current_block;
				}

				mem->size = total_size;
				size_t remaining_size = total_size - size;

				if (remaining_size >= sizeof(struct block_meta) + 8) {
					/* Splitting the remaining block */

					struct block_meta *new_block = (struct block_meta *)((char *)mem + sizeof(struct block_meta) + size);

					new_block->status = STATUS_FREE;
					new_block->size = remaining_size - sizeof(struct block_meta);
					new_block->next = mem->next;
					new_block->prev = mem;

					/* Updating the old memory */
					if (mem->next != NULL)
						mem->next->prev = new_block;

					mem->size = size;
					mem->next = new_block;
				}

				/* The original pointer, resized */
				return ptr;

			} else {
				/* We don't have enough space */

				/* If we are at the end of the heap, we add in it */
				if (((char *)mem + sizeof(struct block_meta) + mem->size == sbrk(0))) {
					size_t need = size - mem->size;

					sbrk(need);
					mem->size += need;

					/* The original pointer, resized bigger (at the end of heap) */
					return ptr;
				}

				/* Canont extend heap, so we malloc */
				void *new_ptr = os_malloc(size);

				if (new_ptr == NULL)
					return NULL;

				size_t copy;

				if (mem->size < size)
					copy = mem->size;
				else
					copy = size;

				/* Copy the old data to the new block */
				memcpy(new_ptr, ptr, copy);
				os_free(ptr);

				return new_ptr;
			}
		}
	}

	if (mem->status == STATUS_MAPPED) {
		/* We will allocate a new block and copy the data */

		void *new_ptr = os_malloc(size);

		if (new_ptr == NULL)
			return NULL;

		size_t copy;

		if (mem->size < size)
			copy = mem->size;
		else
			copy = size;


		memcpy(new_ptr, ptr, copy);
		os_free(ptr);

		return new_ptr;
	}

	/* Returning NULL to not reach control */
	return NULL;
}
